# Original algorithm was published by John Marioni and colleagues as EmptyDrops (Lun, A. et al. Distinguishing cells from empty droplets in droplet-based single-cell RNA sequencing data.)

# This implementation is based on the code in cellranger v3.0 by 10x Genomics

# Copyright 2018 10X Genomics, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import logging

import numpy as np
import scipy.sparse as sparse
import scipy.stats as sp_stats

# Simple Good-Turing estimator.
# Based on S implementation in
# William A. Gale & Geoffrey Sampson (1995) Good-turing frequency estimation without tears,
# Journal of Quantitative Linguistics, 2:3, 217-237, DOI: 10.1080/09296179508590051


class SimpleGoodTuringError(Exception):
	pass


def _averaging_transform(r, nr):
	d = np.concatenate((np.ones(1, dtype=int), np.diff(r)))
	dr = np.concatenate((
		0.5 * (d[1:] + d[0:-1]),
		np.array((d[-1],), dtype=float),
	))
	return nr.astype(float) / dr


def _rstest(r, coef):
	return r * np.power(1 + 1 / r, 1 + coef)


def simple_good_turing(xr, xnr):
	"""Make a Simple Good-Turing estimate of the frequencies.

	Args:
	xr (np.array(int)): Non-zero item frequencies
	xnr (np.array(int)): Non-zero frequencies of frequencies
	Returns:
	(rstar (np.array(float)), p0 (float)):
		rstar: The adjusted non-zero frequencies
		p0: The total probability of unobserved items
	"""

	xr = xr.astype(float)
	xnr = xnr.astype(float)

	xN = np.sum(xr * xnr)

	# Get Linear Good-Turing estimate
	xnrz = _averaging_transform(xr, xnr)
	slope, intercept, _, _, _ = sp_stats.linregress(np.log(xr), np.log(xnrz))

	if slope > -1:
		raise SimpleGoodTuringError("The log-log slope is > -1 (%d); the SGT estimator is not applicable to these data." % slope)

	xrst = _rstest(xr, slope)
	xrstrel = xrst / xr

	# Get traditional Good-Turing estimate
	xrtry = xr == np.concatenate((xr[1:] - 1, np.zeros(1)))
	xrstarel = np.zeros(len(xr))
	xrstarel[xrtry] = (xr[xrtry] + 1) / xr[xrtry] * np.concatenate((xnr[1:], np.zeros(1)))[xrtry] / xnr[xrtry]

	# Determine when to switch from GT to LGT estimates
	tursd = np.ones(len(xr))
	for i in range(len(xr)):
		if xrtry[i]:
			tursd[i] = float(i + 2) / xnr[i] * np.sqrt(xnr[i + 1] * (1 + xnr[i + 1] / xnr[i]))

	xrstcmbrel = np.zeros(len(xr))
	useturing = True
	for r in range(len(xr)):
		if not useturing:
			xrstcmbrel[r] = xrstrel[r]
		else:
			if np.abs(xrstrel[r] - xrstarel[r]) * (1 + r) / tursd[r] > 1.65:
				xrstcmbrel[r] = xrstarel[r]
			else:
				useturing = False
				xrstcmbrel[r] = xrstrel[r]

	# Renormalize the probabilities for observed objects
	sumpraw = np.sum(xrstcmbrel * xr * xnr / xN)

	xrstcmbrel = xrstcmbrel * (1 - xnr[0] / xN) / sumpraw
	p0 = xnr[0] / xN

	return (xr * xrstcmbrel, p0)


def sgt_proportions(frequencies):
	"""Use Simple Good-Turing estimate to adjust for unobserved items

	Args:
	frequencies (np.array(int)): Nonzero frequencies of items
	Returns:
	(pstar (np.array(float)), p0 (float)):
		pstar: The adjusted non-zero proportions
		p0: The total probability of unobserved items
	"""
	if len(frequencies) == 0:
		raise ValueError("Input frequency vector is empty")
	if np.count_nonzero(frequencies) != len(frequencies):
		raise ValueError("Frequencies must be greater than zero")

	freqfreqs = np.bincount(frequencies.astype(np.int64))
	assert freqfreqs[0] == 0
	use_freqs = np.flatnonzero(freqfreqs)

	if len(use_freqs) < 10:
		raise SimpleGoodTuringError("Too few non-zero frequency items (%d). Aborting SGT." % len(use_freqs))

	rstar, p0 = simple_good_turing(use_freqs, freqfreqs[use_freqs])

	# rstar contains the smoothed frequencies.
	# Map each original frequency r to its smoothed rstar.
	rstar_dict = dict(zip(use_freqs, rstar))

	rstar_sum = np.sum(freqfreqs[use_freqs] * rstar)
	rstar_i = np.fromiter((rstar_dict[f] for f in frequencies), dtype=float, count=len(frequencies))
	pstar = (1 - p0) * (rstar_i / rstar_sum)

	assert np.isclose(p0 + np.sum(pstar), 1)
	return (pstar, p0)


def adjust_pvalue_bh(p):
	""" Multiple testing correction of p-values using the Benjamini-Hochberg procedure """
	descending = np.argsort(p)[::-1]
	# q = p * N / k where p = p-value, N = # tests, k = p-value rank
	scale = float(len(p)) / np.arange(len(p), 0, -1)
	q = np.minimum(1, np.minimum.accumulate(scale * p[descending]))

	# Return to original order
	return q[np.argsort(descending)]


def eval_multinomial_loglikelihoods(matrix, profile_p, max_mem_gb=0.1):
	"""Compute the multinomial log PMF for many barcodes
	Args:
	matrix (scipy.sparse.csc_matrix): Matrix of UMI counts (feature x barcode)
	profile_p (np.ndarray(float)): Multinomial probability vector
	max_mem_gb (float): Try to bound memory usage.
	Returns:
	log_likelihoods (np.ndarray(float)): Log-likelihood for each barcode
	"""
	gb_per_bc = float(matrix.shape[0] * matrix.dtype.itemsize) / (1024**3)
	bcs_per_chunk = max(1, int(round(max_mem_gb / gb_per_bc)))
	num_bcs = matrix.shape[1]

	loglk = np.zeros(num_bcs)

	for chunk_start in range(0, num_bcs, bcs_per_chunk):
		chunk = slice(chunk_start, chunk_start + bcs_per_chunk)
		matrix_chunk = matrix[:, chunk].transpose().toarray()
		n = matrix_chunk.sum(1)
		loglk[chunk] = sp_stats.multinomial.logpmf(matrix_chunk, n, p=profile_p)
	return loglk


def simulate_multinomial_loglikelihoods(profile_p, umis_per_bc, num_sims=1000, jump=1000, n_sample_feature_block=1000000, verbose=False):
	"""Simulate draws from a multinomial distribution for various values of N.
	Uses the approximation from Lun et al. ( https://www.biorxiv.org/content/biorxiv/early/2018/04/04/234872.full.pdf )
	Args:
	profile_p (np.ndarray(float)): Probability of observing each feature.
	umis_per_bc (np.ndarray(int)): UMI counts per barcode (multinomial N).
	num_sims (int): Number of simulations per distinct N value.
	jump (int): Vectorize the sampling if the gap between two distinct Ns exceeds this.
	n_sample_feature_block (int): Vectorize this many feature samplings at a time.
	Returns:
	(distinct_ns (np.ndarray(int)), log_likelihoods (np.ndarray(float)):
	distinct_ns is an array containing the distinct N values that were simulated.
	log_likelihoods is a len(distinct_ns) x num_sims matrix containing the
		simulated log likelihoods.
	"""
	distinct_n = np.flatnonzero(np.bincount(umis_per_bc.astype(np.int64)))

	loglk = np.zeros((len(distinct_n), num_sims), dtype=float)

	sampled_features = np.random.choice(len(profile_p), size=n_sample_feature_block, p=profile_p, replace=True)
	k = 0

	log_profile_p = np.log(profile_p)

	for sim_idx in range(num_sims):
		curr_counts = np.ravel(sp_stats.multinomial.rvs(distinct_n[0], profile_p, size=1))

		curr_loglk = sp_stats.multinomial.logpmf(curr_counts, distinct_n[0], p=profile_p)

		loglk[0, sim_idx] = curr_loglk

		for i in range(1, len(distinct_n)):
			step = distinct_n[i] - distinct_n[i - 1]
			if step >= jump:
				# Instead of iterating for each n, sample the intermediate ns all at once
				curr_counts += np.ravel(sp_stats.multinomial.rvs(step, profile_p, size=1))
				curr_loglk = sp_stats.multinomial.logpmf(curr_counts, distinct_n[i], p=profile_p)
				assert not np.isnan(curr_loglk)
			else:
				# Iteratively sample between the two distinct values of n
				for n in range(distinct_n[i - 1] + 1, distinct_n[i] + 1):
					j = sampled_features[k]
					k += 1
					if k >= n_sample_feature_block:
						# Amortize this operation
						sampled_features = np.random.choice(len(profile_p), size=n_sample_feature_block, p=profile_p, replace=True)
						k = 0
					curr_counts[j] += 1
					curr_loglk += log_profile_p[j] + np.log(float(n) / curr_counts[j])

			loglk[i, sim_idx] = curr_loglk

	return distinct_n, loglk


def compute_ambient_pvalues(umis_per_bc, obs_loglk, sim_n, sim_loglk):
	"""Compute p-values for observed multinomial log-likelihoods
	Args:
	umis_per_bc (nd.array(int)): UMI counts per barcode
	obs_loglk (nd.array(float)): Observed log-likelihoods of each barcode deriving from an ambient profile
	sim_n (nd.array(int)): Multinomial N for simulated log-likelihoods
	sim_loglk (nd.array(float)): Simulated log-likelihoods of shape (len(sim_n), num_simulations)
	Returns:
	pvalues (nd.array(float)): p-values
	"""
	assert len(umis_per_bc) == len(obs_loglk)
	assert sim_loglk.shape[0] == len(sim_n)

	# Find the index of the simulated N for each barcode
	sim_n_idx = np.searchsorted(sim_n, umis_per_bc)
	num_sims = sim_loglk.shape[1]

	num_barcodes = len(umis_per_bc)

	pvalues = np.zeros(num_barcodes)

	for i in range(num_barcodes):
		num_lower_loglk = np.sum(sim_loglk[sim_n_idx[i], :] < obs_loglk[i])
		pvalues[i] = float(1 + num_lower_loglk) / (1 + num_sims)
	return pvalues


def estimate_profile_sgt(matrix, barcode_indices, nz_feat):
	""" Estimate a gene expression profile by Simple Good Turing.
	Args:
	raw_mat (sparse matrix): Sparse matrix of all counts
	barcode_indices (np.array(int)): Barcode indices to use
	nz_feat (np.array(int)): Indices of features that are non-zero at least once
	Returns:
	profile (np.array(float)): Estimated probabilities of length len(nz_feat).
	"""
	# Initial profile estimate
	prof_mat = matrix[:, barcode_indices]

	profile = np.ravel(prof_mat[nz_feat, :].sum(axis=1))
	zero_feat = np.flatnonzero(profile == 0)

	# Simple Good Turing estimate
	p_smoothed, p0 = sgt_proportions(profile[np.flatnonzero(profile)])

	# Distribute p0 equally among the zero elements.
	p0_i = p0 / len(zero_feat)

	profile_p = np.repeat(p0_i, len(nz_feat))
	profile_p[np.flatnonzero(profile)] = p_smoothed

	assert np.isclose(profile_p.sum(), 1.0)
	return profile_p


# Construct a background expression profile from barcodes with <= T UMIs
def est_background_profile_sgt(matrix, use_bcs):
	""" Estimate a gene expression profile on a given subset of barcodes.
		Use Good-Turing to smooth the estimated profile.
	Args:
	matrix (scipy.sparse.csc_matrix): Sparse matrix of all counts
	use_bcs (np.array(int)): Indices of barcodes to use (col indices into matrix)
	Returns:
	profile (use_features, np.array(float)): Estimated probabilities of length use_features.
	"""
	# Use features that are nonzero anywhere in the data
	use_feats = np.flatnonzero(np.asarray(matrix.sum(1)))

	# Estimate background profile
	bg_profile_p = estimate_profile_sgt(matrix, use_bcs, use_feats)

	return (use_feats, bg_profile_p)


# Sten Linnarsson's version (Aug 2019)
def call_cells(matrix: sparse.csr_matrix, expected_n_cells: int = 5000) -> np.ndarray:
	"""
	Determine likely true cells among the barcodes by contrasting with the ambient RNA profile

	Args:
		matrix: 			expression matrix
		expected_n_cells:	expected number of true cells in the sample

	Returns:
		calls:	vector of bools indicating true cell barcodes
	"""
	n_barcodes = matrix.shape[1]
	expected_n_cells = min(expected_n_cells, n_barcodes // 5)
	total_umis = np.array(matrix.sum(axis=0))[0]  # total UMIs per barcode
	# upper limit of UMIs for barcodes considered ambient, calculated as greatest UMI count after removing twice the expected number of cells
	max_ambient_umis = np.percentile(total_umis, 100 * (n_barcodes - expected_n_cells * 2) / n_barcodes)
	# median number of UMIs among the top expected_n_cells barcodes
	median_initial_umis = np.median(total_umis[total_umis > np.percentile(total_umis, 100 * (n_barcodes - expected_n_cells) / n_barcodes)])
	min_cell_umis = int(max(500, median_initial_umis * 0.1))  # 10% of median, but at least 500 UMIs

	# Ambient RNA beads, covering the range 20 to max_amient_umis
	ambient_bcs = (total_umis < max_ambient_umis) & (total_umis > 20)
	if ambient_bcs.sum() == 0:
		# No beads were ambient, because cells had very low UMIs
		logging.warning("No ambient RNA beads were found; maybe sample had too few cells?")
		return max_ambient_umis, np.ones_like(total_umis)
	try:
		eval_features, ambient_profile_p = est_background_profile_sgt(matrix, ambient_bcs)
	except SimpleGoodTuringError as e:
		logging.error(e)
		return max_ambient_umis, np.ones_like(total_umis)

	# Evaluate candidate barcodes
	eval_bcs = total_umis > min_cell_umis
	eval_mat = matrix[eval_features, :][:, eval_bcs]

	# Compute observed log-likelihood of barcodes being generated from ambient RNA
	obs_loglk = eval_multinomial_loglikelihoods(eval_mat, ambient_profile_p)

	# Simulate log likelihoods
	distinct_ns, sim_loglk = simulate_multinomial_loglikelihoods(ambient_profile_p, total_umis[eval_bcs], num_sims=1000, verbose=True)

	# Compute p-values
	pvalues = compute_ambient_pvalues(total_umis[eval_bcs], obs_loglk, distinct_ns, sim_loglk)
	pvalues_adj = adjust_pvalue_bh(pvalues)
	pvalues_adj_all = np.ones_like(total_umis)
	pvalues_adj_all[eval_bcs] = pvalues_adj
	return max_ambient_umis, pvalues_adj_all



from .utils import get_loom_spec_version, compare_loom_spec_version, timestamp, deprecated
from .normalize import normalize_attr_values, materialize_attr_values
from .attribute_manager import AttributeManager
from .global_attribute_manager import GlobalAttributeManager
from .graph_manager import GraphManager
from .layer_manager import LayerManager
from .loom_view import LoomView
from .loom_layer import MemoryLoomLayer, LoomLayer
from .to_html import to_html
from .view_manager import ViewManager
from .loompy import connect, create, create_append, combine, create_from_cellranger, LoomConnection, new, combine_faster, create_from_matrix_market, create_from_star, create_from_tsv
from .loom_validator import LoomValidator
from ._version import __version__, loom_spec_version
from .bus_file import create_from_fastq
from .cell_calling import call_cells
from .metadata_loaders import load_gene_metadata, make_row_attrs_from_gene_annotations, make_row_attrs_from_gene_metadata, load_sample_metadata



import numpy as np
import loompy
from typing import *


class ViewManager:
	"""
	Create views by slicing an underlying LoomConnection or LoomView
	"""
	def __init__(self, ds: Any) -> None:
		self.ds = ds

	def __getitem__(self, slice_: Tuple[Union[slice, np.ndarray, int], Union[slice, np.ndarray, int]]) -> loompy.LoomView:
		"""
		Create a new view by slicing through the loom file or view

		Args:
			slice_ (2-tuple of slice, int or np.ndarray): 	How to slice the file or view

		Returns:
			A LoomView object, an in-memory representation of the sliced file
		"""
		if type(slice_) is not tuple or len(slice_) != 2:
			raise ValueError("Views require slices along two dimensions")

		rows = slice_[0]
		cols = slice_[1]

		ra = self.ds.ra[rows]
		row_graphs = self.ds.row_graphs[rows]
		ca = self.ds.ca[cols]
		col_graphs = self.ds.col_graphs[cols]
		layers = self.ds.layer[rows, cols]

		return loompy.LoomView(layers, ra, ca, row_graphs, col_graphs, filename=self.ds.filename, file_attrs=self.ds.attrs)



import numpy as np
import scipy.sparse as sparse
from typing import *
import html
import logging


def normalize_attr_strings(a: np.ndarray) -> np.ndarray:
	"""
	Take an np.ndarray of all kinds of string-like elements, and return an array of ascii (np.bytes_) objects
	"""
	if np.issubdtype(a.dtype, np.object_):
		# if np.all([type(x) is str for x in a]) or np.all([type(x) is np.str_ for x in a]):
		if np.all([(type(x) is str or type(x) is np.str_) for x in a]):
			return np.array([x.encode('ascii', 'xmlcharrefreplace') for x in a])
		elif np.all([type(x) is np.bytes_ for x in a]):
			return a.astype("bytes_")
		else:
			logging.debug(f"Attribute contains mixed object types ({np.unique([str(type(x)) for x in a])}); casting all to string")
			return np.array([str(x) for x in a], dtype="bytes_")
	elif np.issubdtype(a.dtype, np.bytes_) or np.issubdtype(a.dtype, np.object_):
		return a
	elif np.issubdtype(a.dtype, np.str_):
		return np.array([x.encode('ascii', 'xmlcharrefreplace') for x in a])
	else:
		raise ValueError("String values must be object, ascii or unicode.")


def normalize_attr_array(a: Any) -> np.ndarray:
	"""
	Take all kinds of array-like inputs and normalize to a one-dimensional np.ndarray
	"""
	if type(a) is np.ndarray:
		return a
	elif type(a) is np.matrix:
		if a.shape[0] == 1:
			return np.array(a)[0, :]
		elif a.shape[1] == 1:
			return np.array(a)[:, 0]
		else:
			raise ValueError("Attribute values must be 1-dimensional.")
	elif type(a) is list or type(a) is tuple:
		return np.array(a)
	elif sparse.issparse(a):
		return normalize_attr_array(a.todense())
	else:
		raise ValueError("Argument must be a list, tuple, numpy matrix, numpy ndarray or sparse matrix.")


def normalize_attr_values(a: Any, use_object_strings: bool = False) -> np.ndarray:
	"""
	Take all kinds of input values and validate/normalize them.
	
	Args:
		a	List, tuple, np.matrix, np.ndarray or sparse matrix
			Elements can be strings, numbers or bools
	
	Returns
		a_normalized    An np.ndarray with elements conforming to one of the valid Loom attribute types
	
	Remarks:
		This method should be used to prepare the values to be stored in the HDF5 file. You should not
		return the values to the caller; for that, use materialize_attr_values()
	"""
	scalar = False
	if np.isscalar(a):
		a = np.array([a])
		scalar = True
	arr = normalize_attr_array(a)
	if np.issubdtype(arr.dtype, np.integer) or np.issubdtype(arr.dtype, np.floating):
		pass  # We allow all these types
	elif np.issubdtype(arr.dtype, np.character) or np.issubdtype(arr.dtype, np.object_):
		if use_object_strings:
			arr = np.array([str(elm) for elm in a], dtype=object)
		else:
			arr = normalize_attr_strings(arr)
	elif np.issubdtype(arr.dtype, np.bool_):
		arr = arr.astype('ubyte')
	if scalar:
		return arr[0]
	else:
		return arr


def materialize_attr_values(a: np.ndarray) -> np.ndarray:
	scalar = False
	if np.isscalar(a):
		scalar = True
		a = np.array([a])
	result: np.ndarray = None   # This second clause takes care of attributes stored as variable-length ascii, which can be generated by loomR or Seurat
	if np.issubdtype(a.dtype, np.bytes_) or np.issubdtype(a.dtype, np.object_):
		# First ensure that what we load is valid ascii (i.e. ignore anything outside 7-bit range)
		if len(a) > 0 and hasattr(a[0], "decode"):  # This takes care of Loom files that store strings as UTF8, which comes in as str and doesn't have a decode method
			temp = np.array([x.decode('ascii', 'ignore') for x in a])
		else:
			temp = a
		# Then unescape XML entities and convert to unicode
		try:
			result = np.array([html.unescape(x) for x in temp.astype(str)], dtype=object)
		except:  # Dirty hack to handle UTF-8 non-break-space in scalar strings. TODO: Rewrite this whole method completely!
			if type(a[0]) == np.bytes_:
				result = [ a[0].replace(b'\xc2\xa0', b'') ]
	elif np.issubdtype(a.dtype, np.str_):
		result = np.array(a.astype(str), dtype=object)
	else:
		result = a
	if scalar:
		return result[0]
	else:
		return result




from collections import defaultdict

d = "./"

mgiID2MRK_ENSE = {}
enseID2mgiID = {}
with open(d + "inputs/MRK_ENSEMBL.rpt") as f:
	for line in f:
		items = line[:-1].split("\t")
		mgiID = items[0]
		mgiID2MRK_ENSE[mgiID] = items
		enseID = items[5]
		enseID2mgiID[enseID] = mgiID

enseID2MGI_GMC = {}
with open(d + "inputs/MGI_Gene_Model_Coord.rpt") as f:
	MGI_GMC_headers = f.readline()[:-1].split("\t")
	for line in f:
		items = line[:-1].split("\t")
		enseID = items[10]
		enseID2MGI_GMC[enseID] = items

mgiID2MRK_Seq = {}
with open(d + "inputs/MRK_Sequence.rpt") as f:
        MRK_Seq_headers = f.readline()[:-1].split("\t")
        for line in f:
                items = line[:-1].split("\t")
                mgiID = items[0]
                mgiID2MRK_Seq[mgiID] = items

enseID2mart = {}
with open(d + "inputs/mart_export.txt") as f:
	mart_headers = f.readline()[:-1].split("\t")
	for line in f:
		items = line[:-1].split("\t")
		enseID = items[0]
		enseID2mart[enseID] = items

geneSymbol2TF = {}
with open(d + "inputs/TF_TcoF-DB.tsv") as f:
	TF_headers = f.readline()[:-1].split("\t")
	for line in f:
		items = line[:-1].split("\t")
		geneSymbol = items[0]
		geneSymbol2TF[geneSymbol] = items

geneSymbol2Regulated = defaultdict(list)
with open(d + "inputs/trrust_rawdata.mouse.tsv") as f:
	for line in f:
		items = line[:-1].split("\t")
		TFSymbol = items[0]
		geneSymbol2Regulated[TFSymbol].append(items[1])

with open(d + "gencode.vM23.metadata.tab", "w") as fout:
	fout.write("\t".join([
		"Accession",
		"AccessionVersion",
		"Gene",
		"FullName",
		"GeneType",
		"HgncID",
		"Chromosome",
		"Strand",
		"ChromosomeStart",
		"ChromosomeEnd",
		"LocusGroup",
		"LocusType",
		"Location",
		"LocationSortable",
		"Aliases",
		"VegaID",
		"UcscID",
		"RefseqID",
		"CcdsID",
		"UniprotID",
		"PubmedID",
		"MgdID",
		"RgdID",
		"CosmicID",
		"OmimID",
		"MirBaseID",
		"IsTFi (TcoF-DB)",
		"DnaBindingDomain",
		"Regulates (TRRUST)"
	]))
	fout.write("\n")
	with open(d + "inputs/gencode.vM23.primary_assembly.annotation.gtf") as f:
		for line in f:
			if line.startswith("##"):
				continue
			items = line[:-1].split("\t")
			if items[2] != "gene":
				continue
			extra = {x.strip().split(" ")[0]: x.strip().split(" ")[1].strip('"') for x in items[8].split(";")[:-1]}
			enseID = extra["gene_id"].split(".")[0]
			geneSymbol = extra.get("gene_name", "")
			fout.write("\t".join([
				enseID,
				extra["gene_id"],
				geneSymbol,
				enseID2MGI_GMC[enseID][3] if enseID in enseID2MGI_GMC else "",  # full name
				extra["gene_type"],  # gene type from gencode
				"",  # HGNC id
				items[0],  # Chromosome
				items[6],
				items[3],  # Start
				items[4],  # End
				"",  # Locus group
				mgiID2MRK_ENSE[mgiID][8],  # Locus type
				"",  # Location
				"",  # Location, sortable
				"",  # Aliases
				enseID2mart[enseID][5] if enseID in enseID2mart else "",  # VEGA id
				enseID2mart[enseID][4] if enseID in enseID2mart else "",  # UCSC id
				mgiID2MRK_Seq[mgiID][12],  # Refseq id
				enseID2mart[enseID][6] if enseID in enseID2mart else "",  # CCDS id
				mgiID2MRK_Seq[mgiID][14],  # Uniprot id
				"",  # Pubmed id
				"",  # MGD id
				"",  # RGD id
				"",  # COSMIC id
				"",  # OMIM id
				"",  # MIRbase id
				"True" if (geneSymbol in geneSymbol2TF) else "False", #  IsTF?
				"", # DBD
				",".join(geneSymbol2Regulated[geneSymbol])  #  TF regulated genes
				]))
			fout.write("\n")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# loompy documentation build configuration file, created by
# sphinx-quickstart on Tue Oct  3 00:11:17 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
import sphinx_bootstrap_theme
sys.path.insert(0, os.path.abspath('..'))
sys.path.insert(0, os.path.abspath('.'))
sys.path.insert(0, os.path.abspath('_ext'))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',
    'sphinx.ext.napoleon',
    'sphinx.ext.intersphinx',
    'edit_on_github']

intersphinx_mapping = {
    'urllib3': ('http://urllib3.readthedocs.org/en/latest', None),
    'python': ('http://docs.python.org/3', None),
    'numpy': ('http://docs.scipy.org/doc/numpy/', None),
    'scipy': ('http://docs.scipy.org/doc/scipy/reference', None),
    'matplotlib': ('http://matplotlib.org', None),
    'h5py': ('http://docs.h5py.org/en/latest/', None)
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'loompy'
copyright = '2017, LinnarssonLab'
author = 'Linnarsson lab members'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#

__version__ = "0.0.0"
exec(open('../loompy/_version.py').read())
# The short X.Y version.
version = ".".join(__version__.split(".")[:-1])
# The full version, including alpha/beta/rc tags.
release = __version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = 'sphinx_rtd_theme'
html_theme = 'bootstrap'
html_theme_path = sphinx_bootstrap_theme.get_html_theme_path()

# (Optional) Logo. Should be small enough to fit the navbar (ideally 24x24).
# Path should be relative to the ``_static`` files directory.
html_logo = "Loom_icon.png"

# Theme options are theme-specific and customize the look and feel of a
# theme further.
html_theme_options = {
    # Navigation bar title. (Default: ``project`` value)
    'navbar_title': "Loom",

    # Tab name for entire site. (Default: "Site")
    'navbar_site_name': "Contents",

    # A list of tuples containing pages or urls to link to.
    # Valid tuples should be in the following forms:
    #    (name, page)                 # a link to a page
    #    (name, "/aa/bb", 1)          # a link to an arbitrary relative url
    #    (name, "http://example.com", True) # arbitrary absolute url
    # Note the "1" or "True" value above as the third argument to indicate
    # an arbitrary url.
    'navbar_links': [
        ("loompy.org", "http://loompy.org", True),
        ("GitHub", "https://github.com/linnarsson-lab/loompy", True),
    ],

    # Render the next and previous page links in navbar. (Default: true)
    'navbar_sidebarrel': False,

    # Render the current pages TOC in the navbar. (Default: true)
    'navbar_pagenav': False,

    # Tab name for the current pages TOC. (Default: "Page")
    'navbar_pagenav_name': "Page",

    # Global TOC depth for "site" navbar tab. (Default: 1)
    # Switching to -1 shows all levels.
    'globaltoc_depth': 2,

    # Include hidden TOCs in Site navbar?
    #
    # Note: If this is "false", you cannot have mixed ``:hidden:`` and
    # non-hidden ``toctree`` directives in the same page, or else the build
    # will break.
    #
    # Values: "true" (default) or "false"
    'globaltoc_includehidden': "true",

    # HTML navbar class (Default: "navbar") to attach to <div> element.
    # For black navbar, do "navbar navbar-inverse"
    'navbar_class': "navbar navbar-light",

    # Fix navigation bar to top of page?
    # Values: "true" (default) or "false"
    'navbar_fixed_top': "true",

    # Location of link to source.
    # Options are "nav" (default), "footer" or anything else to exclude.
    'source_link_position': "",

    # Bootswatch (http://bootswatch.com/) theme.
    #
    # Options are nothing (default) or the name of a valid theme
    # such as "cosmo" or "sandstone".
    #
    # The set of valid themes depend on the version of Bootstrap
    # that's used (the next config option).
    #
    # Currently, the supported themes are:
    # - Bootstrap 2: https://bootswatch.com/2
    # - Bootstrap 3: https://bootswatch.com/3
    'bootswatch_theme': "simplex",

    # Choose Bootstrap version.
    # Values: "3" (default) or "2" (in quotes)
    'bootstrap_version': "3",

}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    '**': ['localtoc.html']
}


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'loompydoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'loompy.tex', 'loompy Documentation',
     'LinnarssonLab', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'loompy', 'loompy Documentation',
     [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'loompy', 'loompy Documentation',
     author, 'loompy', 'One line description of project.',
     'Miscellaneous'),
]



# -- Options for Epub output ----------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project
epub_author = author
epub_publisher = author
epub_copyright = copyright

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']

# --- Options for "Edit on GitHub" ---
edit_on_github_project = 'linnarsson-lab/loompy'
edit_on_github_branch = 'master/doc'

# --- Sort the autodocs by source (currently not workin not sure why)
autodoc_member_order = 'bysource'

html_scaled_image_link = False


import sys, os

extent = 600  # how many bases away from polya to include
min_len = 90  # how many non-repeat bases required to make a transcript

from typing import *
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq

def find_polys(seq: SeqRecord, c: str = "A", n: int = 15) -> List[Tuple[int, int]]:
	found = []
	count = seq[:n].count(c)  # Count occurences in the first k-mer
	if count >= n - 1:  # We have a match
		found.append(0)
	ix = 0
	while ix < len(seq) - n - 1:
		if seq[ix] == c:  # Outgoing base
			count -= 1
		if seq[ix + n] == c:  # Incoming base
			count += 1
		ix += 1
		if count >= n - 1:  # We have a match
			found.append(ix)
	
	sorted_by_lower_bound = [(f, f + n) for f in found]
	# merge intervals (https://codereview.stackexchange.com/questions/69242/merging-overlapping-intervals)
	merged = []
	for higher in sorted_by_lower_bound:
		if not merged:
			merged.append(higher)
		else:
			lower = merged[-1]
			# test for intersection between lower and higher:
			# we know via sorting that lower[0] <= higher[0]
			if higher[0] <= lower[1]:
				upper_bound = max(lower[1], higher[1])
				merged[-1] = (lower[0], upper_bound)  # replace by merged interval
			else:
				merged.append(higher)
	return merged

polyAs = {}
polyTs = {}
for fasta in SeqIO.parse(open("inputs/gencode.vM23.unspliced.fa"),'fasta'):
	gene_id = fasta.id
	intervals = find_polys(fasta.seq, c="A", n=14)
	if len(intervals) > 0:
		polyAs[gene_id] = intervals
	# Collect fragments on the opposite strand, downstream of poly-Ts (not sure if such reads really happen?)
	intervals = find_polys(fasta.seq, c="T", n=14)
	if len(intervals) > 0:
		polyTs[gene_id] = intervals

tr2g = {}
with open("inputs/gencode.vM23.primary_assembly.annotation.gtf") as f:
	for line in f:
		if "\ttranscript\t" in line:
			items = line.split("; ")
			chrom, _, _, start, end, _, strand, _, gid = items[0].split("\t")
			gene_id = gid.split('"')[1]
			transcript_id = items[1].split('"')[1]
			gene_type = items[2].split('"')[1]
			gene_name = items[3].split('"')[1]
			tr2g[transcript_id] = (chrom, start, end, strand, gene_id, gene_type, gene_name)

count = 0
with open("fragments2genes.txt", "w") as ftr2g:
	with open("inputs/gencode.vM23.fragments.fa", "w") as fout:
		# Write the nascent fragments, with one partial transcript per internal poly-A/T site
		with open("unspliced_fragments.txt", "w") as fucapture:
			for fasta in SeqIO.parse(open("inputs/gencode.vM23.unspliced.fa"),'fasta'):  # Note we're in the masked file now
				gene_id = fasta.id
				if gene_id in polyAs:
					for interval in polyAs[gene_id]:
						seq = str(fasta.seq[max(0, interval[0] - extent):interval[0]])
						#seq = seq.translate(tr).strip("N")
						if len(seq) >= min_len:
							count += 1
							transcript_id = f"{gene_id}.A{interval[0]}"
							trseq = SeqRecord(Seq(seq), transcript_id, '', '')
							fout.write(trseq.format("fasta"))
							ftr2g.write(f"{transcript_id}\t{gene_id}\n")
							fucapture.write(f"{transcript_id}\n")
				if gene_id in polyTs:
					for interval in polyTs[gene_id]:
						seq = str(fasta.seq[interval[1]:interval[1] + extent].reverse_complement())
						#seq = seq.translate(tr).strip("N")
						if len(seq) >= min_len:
							count += 1
							transcript_id = f"{gene_id}.T{interval[0]}"
							trseq = SeqRecord(Seq(seq), transcript_id, '', '')
							fout.write(trseq.format("fasta"))
							ftr2g.write(f"{transcript_id}\t{gene_id}\n")
							fucapture.write(f"{transcript_id}\n")
		# Write the mature fragments, covering the 3' end of each mature transcript
		with open("spliced_fragments.txt", "w") as fscapture:
			for fasta in SeqIO.parse(open("inputs/gencode.vM23.transcripts.fa"),'fasta'):  # Note we're in the masked file now
				transcript_id = fasta.id.split("|")[0]
				gene_id = fasta.id.split("|")[1]
				attrs = tr2g[transcript_id]
				seq = str(fasta.seq[-extent:])
				if len(seq) >= min_len:
					count += 1
					trseq = SeqRecord(Seq(seq), f"{transcript_id}.{count} gene_id:{attrs[4]} gene_name:{attrs[6]}", '', '')
					fout.write(trseq.format("fasta"))
					ftr2g.write(f"{transcript_id}.{count}\t{attrs[4]}\n")
					fscapture.write(f"{transcript_id}.{count}\n")




from typing import Dict, Union, Iterable
import os.path, re
import sqlite3 as sqlite
import numpy as np

def load_gene_metadata(gtf_file : str) -> Dict[str, Dict[str, Union[int, str]]]:
	"""
	Read gene metadata from a GTF file.

	Args:
	  gtf_file (str):             path to GTF file

	Returns:
	  A Dict with each Accession (gtf "gene_id") pointing to a Dict of metadata keys -> values
	"""
	if not os.path.exists(gtf_file):
		raise ValueError(f"Gene metadata file '{gtf_file}' not found.")
	regex_genetype = re.compile('gene_biotype "([^"]+)"')
	regex_geneid = re.compile('gene_id "([^"]+)"')
	regex_genename = re.compile('gene_name "([^"]+)"')
	geneid2annots = {}
	for line in open(gtf_file).readlines():
		if line.startswith('#'):
			continue
		fields = line.rstrip().split('\t')
		chrom, feature_class, feature_type, start_str, end_str, junk, strand, junk, tags = fields
		if feature_type == "gene":
			genename = geneid = regex_geneid.search(tags).group(1)
			_genename_search = regex_genename.search(tags)
			if _genename_search:
				genename = _genename_search.group(1)
			_genetype_search = regex_genetype.search(tags)
			genetype = _genetype_search.group(1) if _genetype_search else "n/a"
			chrid, start, end = fields[0], int(fields[3]), int(fields[4])
			attrs = { "Gene": genename, "Accession": geneid, "Biotype": genetype, \
                                                  "Chromosome": chrid, "Start": start, "End": end }
			geneid2annots[geneid] = attrs
			geneid2annots[genename] = attrs
	return geneid2annots

def make_row_attrs_from_gene_annotations(acc2annots : Dict[str, Dict[str, Union[int, str]]], ordered_features : Iterable[str]) -> Dict[str, np.ndarray]:
	"""
         Construct loom row attributes corresponding to ordered_features from load_gene_metadata output.

	Args:
	  acc2annots (Dict):          output from load_gene_metadata
          ordered_features (str):     the accessions (should match the gtf "gene_id") in matrix row order

	Returns:
          A row attribute dictionary of attr->numpy arrays to assign to a Loom object.
	"""
	ra = {}
	first_annot = next(iter(acc2annots.values()))
	ra_attrs = list(first_annot.keys())
	n_genes = len(ordered_features)
	for ra_attr in ra_attrs:
		ra[ra_attr] = np.zeros((n_genes,), dtype = object)
	for idx, geneid in enumerate(ordered_features):
		try:
			annots = acc2annots[geneid]
		except KeyError:
			annots = { "Gene": geneid, "Accession": geneid, "Biotype": "n/a", "Chromosome": "Un", "Start": "0", "End": "0" }
		for ra_attr in ra_attrs:
			ra[ra_attr][idx] = annots[ra_attr]
	return ra

def make_row_attrs_from_gene_metadata(gtf_file : str, ordered_features : Iterable[str]) -> Dict[str, np.ndarray]:
	"""
        Read gene metadata from a GTF file and construct loom row attributes corresponding to ordered_features.

	Args:
	  gtf_file (str):             path to GTF file
          ordered_features (str):     the accessions (should match the gtf "gene_id") in matrix row order

	Returns:
          A row attribute object ready to assign to a Loom object.
	"""
	acc2annots = load_gene_metadata(gtf_file)
	return make_row_attrs_from_gene_annotations(acc2annots, ordered_features)

def load_sample_metadata(path: str, sample_id: str) -> Dict[str, str]:
        """
        Read sample metadata from either an sqlite '.db' database or a tab-delimited file.
        The tab-file has to have  one sample per line and a header with a column
        labelled 'Name' or 'SampleId' in which a match to function argument sample_id should be found.

        Args:
          path (str):             path to sqlite database (.db) or tab-delimited metadata file
          sample_id (str):        Id of sample to annotate

        Returns:
              A Dict of metadata keys -> values
        """
        if not os.path.exists(path):
        	raise ValueError(f"Samples metadata file '{path}' not found.")
        if path.endswith(".db"):
                # sqlite3
                with sqlite.connect(path) as db:
                        cursor = db.cursor()
                        cursor.execute("SELECT * FROM sample WHERE name = ?", (sample_id,))
                        keys = [x[0] for x in cursor.description]
                        vals = cursor.fetchone()
                        if vals is not None:
                                return dict(zip(keys, vals))
                        raise ValueError(f"SampleID '{sample_id}' was not found in the samples database.")
        else:
                result = {}
                with open(path) as f:
                        headers = [x.lower() for x in f.readline()[:-1].split("\t")]
                        if "sampleid" not in headers and 'name' not in headers:
                                raise ValueError("Required column 'SampleID' or 'Name' not found in sample metadata file")
                        if "sampleid" in headers:
                                sample_metadata_key_idx = headers.index("sampleid")
                        else:
                                sample_metadata_key_idx = headers.index("name")
                        sample_found = False
                        for line in f:
                                items = line[:-1].split("\t")
                                if len(items) > sample_metadata_key_idx and items[sample_metadata_key_idx] == sample_id:
                                        for i, item in enumerate(items):
                                                result[headers[i]] = item
                                        sample_found = True
                if not sample_found:
                        raise ValueError(f"SampleID '{sample_id}' not found in sample metadata file")
                return result




from setuptools import find_packages, setup

# First update the version in loompy/_version.py, then:

# cd ~/code/loompy  (the directory where loompy resides)
# rm -r dist   (otherwise twine will upload the oldest build!)
# python setup.py sdist
# twine upload dist/*

# NOTE: Don't forget to update the release version at loompy.github.io (index.html)!

# pylint: disable=exec-used
__version__ = '0.0.0'
exec(open('loompy/_version.py').read())

setup(
	name="loompy",
	version=__version__,
	packages=find_packages(),
	python_requires='>=3.6',
	install_requires=['h5py', 'numpy', 'scipy', 'setuptools', 'numba', 'click', "numpy-groupies"],
	entry_points='''
		[console_scripts]
		loompy=loompy.commands:cli
	''',
	# metadata for upload to PyPI
	author="Linnarsson Lab",
	author_email="sten.linnarsson@ki.se",
	description="Work with Loom files for single-cell RNA-seq data",
	license="BSD",
	keywords="loom omics transcriptomics bioinformatics",
	url="https://github.com/linnarsson-lab/loompy",
	download_url=f"https://github.com/linnarsson-lab/loompy/archive/{__version__}.tar.gz",
)



from typing import *
import numpy as np
import loompy
import logging
import scipy.sparse as sparse


class LayerManager:
	"""
	Manage a set of layers with a backing HDF5 file store
	"""
	def __init__(self, ds: Any) -> None:  # Note: can't give type for ds because it will be circular and mypy doesn't support it
		"""
		Create a LayerManager object.
		"""
		setattr(self, "!ds", ds)
		storage: Dict[str, np.ndarray] = {}
		setattr(self, "!storage", storage)
		if ds is not None:
			if "matrix" in ds._file:
				self.__dict__["storage"][""] = None
			if "layers" in ds._file:
				for key in self.ds._file["layers"].keys():
					self.__dict__["storage"][key] = None
			elif ds.mode == "r+":
				ds._file.create_group('/layers')

	def last_modified(self, name: str = None) -> str:
		"""
		Return a compact ISO8601 timestamp (UTC timezone) indicating when the layer was last modified

		Note: if name is None, the modification time of the most recently modified layer is returned
		"""
		if name is not None:
			return self[name].last_modified()
		ts = ""
		for name in self.keys():
			if ts is None:
				ts = self[name].last_modified()
			else:
				if self[name].last_modified() > ts:
					ts = self[name].last_modified()
		return ts

	def keys(self) -> List[str]:
		return list(self.__dict__["storage"].keys())

	def items(self) -> Iterable[Tuple[str, np.ndarray]]:
		for key in self.keys():
			yield (key, self[key])

	def __len__(self) -> int:
		return len(self.keys())

	def __contains__(self, name: str) -> bool:
		return name in self.keys()

	def __iter__(self) -> Iterator[str]:
		for key in self.keys():
			yield key

	def __getitem__(self, thing: Any) -> np.ndarray:
		"""
		Access a layer by name, or slice through all the layers

		Args:
			thing:		if string, return the specified layer ("" is the default layer)
						if slice 2-tuple, return a new LayerManager with all layers sliced
		"""
		if type(thing) is str:
			return self.__getattr__(thing)
		else:
			# Assume some kind of slice
			lm = LayerManager(None)
			for key, layer in self.items():
				lm[key] = loompy.MemoryLoomLayer(key, layer[thing])
			return lm

	def __getattr__(self, name: str) -> np.ndarray:
		try:
			vals = self.__dict__["storage"][name]
			if vals is None:
				# Read values from the HDF5 file
				return loompy.LoomLayer(name, self.ds)
			return vals
		except KeyError:
			raise AttributeError(f"'{type(self)}' object has no attribute '{name}'")

	def __setitem__(self, name: str, val: np.ndarray) -> None:
		return self.__setattr__(name, val)

	def __setattr__(self, name: str, val: np.ndarray) -> None:
		if name.startswith("!"):
			super(LayerManager, self).__setattr__(name[1:], val)
		elif "/" in name:
			raise KeyError("Layer name cannot contain slash (/)")
		else:
			if self.ds is not None:
				if type(val) is str:  # val specifies the dtype of an empty layer
					matrix: np.ndarray = None
					dtype = val
					shape = self.ds.shape
				elif sparse.issparse(val):  # val is a sparse matrix
					matrix = None
					dtype = val.dtype
					shape = val.shape
				else:  # val is a matrix that will be the layer
					matrix = val
					dtype = matrix.dtype
					shape = matrix.shape
					if not np.isfinite(matrix).all():
						raise ValueError("INF and NaN not allowed in loom matrix")
				if name != "" and shape != self.ds.shape:
						raise ValueError(f"All layers must have same shape {self.ds.shape}")
				if self.ds._file.mode != "r+":
					raise IOError("Cannot save layers when connected in read-only mode")
				if not (np.issubdtype(dtype, np.integer) or np.issubdtype(dtype, np.floating)):
					raise ValueError("Matrix elements must be integer or float")
				if not self.ds._file.__contains__("/layers"):
					self.ds._file.create_group("/layers")

				# make sure chunk size is not bigger than actual matrix size
				chunks = (min(64, shape[0]), min(64, shape[1]))
				path = "/layers/" + name
				if name == "":
					path = "/matrix"
				if self.ds._file.__contains__(path):
					del self.ds._file[path]

				# Save the matrix
				self.ds._file.create_dataset(
					path,
					data=matrix,
					dtype=dtype,
					shape=shape,
					maxshape=(shape[0], None),
					chunks=chunks,
					fletcher32=False,
					compression="gzip",
					shuffle=False,
					compression_opts=2
				)
				if name == "":
					self.ds.shape = shape
				self.__dict__["storage"][name] = None

				# Fill the matrix with sparse data
				if sparse.issparse(val):
					m = val.tocsc()
					window = max(1, 1024**3 // 8 * m.shape[0])
					ix = 0
					while ix < val.shape[1]:
						window = min(window, m.shape[1] - ix)
						if window == 0:
							break
						self.ds._file[path][:, ix:ix + window] = m[:, ix: ix + window].toarray()
						ix += window

				self.ds._file.flush()
			else:
				self.__dict__["storage"][name] = val

	def __delitem__(self, name: str) -> None:
		return self.__delattr__(name)

	def __delattr__(self, name: str) -> None:
		if self.ds is not None:
			if name == "":
				raise ValueError("Cannot delete default layer")
			else:
				path = "/layers/" + name
				if self.ds._file.__contains__(path):
					del self.ds._file[path]
				self.ds._file.flush()
		else:
			if name in self.__dict__["storage"]:
				del self.__dict__["storage"][name]

	def _permute(self, ordering: np.ndarray, *, axis: int) -> None:
		for key in self.keys():
			self[key]._permute(ordering, axis=axis)



import h5py
from typing import *
import logging
import numpy as np
import loompy

from .utils import get_loom_spec_version


class LoomValidator:
	def __init__(self, version: str = None) -> None:
		"""
		Args:
			version: 		The Loom file format version to validate against ("3.0.0", "2.0.1", "old"), or None to infer from file
		
		Remarks:
			"old" version will accept files that lack the "row_graphs" and "col_graphs" groups
		"""
		self.version = version  #: Version of the spec to validate against
		self.errors: List[str] = []  #: Errors found during validation
		self.warnings: List[str] = []  #: Warnings triggered during validation
		self.summary: List[str] = []  #: Summary of the file structure

	def _check(self, condition: bool, message: str) -> bool:
		if not condition:
			self.errors.append(message)
		return condition
	
	def _warn(self, condition: bool, message: str) -> bool:
		if not condition:
			self.warnings.append(message)
		return condition

	def validate(self, path: str, strictness: str = "speconly") -> bool:
		"""
		Validate a file for conformance to the Loom specification

		Args:
			path: 			Full path to the file to be validated
			strictness:		"speconly" or "conventions"

		Remarks:
			In "speconly" mode, conformance is assessed relative to the file format specification
			at http://linnarssonlab.org/loompy/format/. In "conventions" mode, conformance is additionally
			assessed relative to attribute name and data type conventions given at http://linnarssonlab.org/loompy/conventions/.
		"""
		valid1 = True
		with h5py.File(path, mode="r") as f:
			if self.version == None:
				self.version = get_loom_spec_version(f)
			valid1 = self.validate_spec(f)
			if not valid1:
				self.errors.append("For help, see http://linnarssonlab.org/loompy/format/")

		valid2 = True
		if strictness == "conventions":
			with loompy.connect(path, mode="r") as ds:
				valid2 = self.validate_conventions(ds)
				if not valid2:
					self.errors.append("For help, see http://linnarssonlab.org/loompy/conventions/")

		return valid1 and valid2

	def validate_conventions(self, ds: loompy.LoomConnection) -> bool:
		"""
		Validate the LoomConnection object against the attribute name/dtype conventions.

		Args:
			ds:			LoomConnection object
		
		Returns:
			True if the file conforms to the conventions, else False
		
		Remarks:
			Upon return, the instance attributes 'self.errors' and 'self.warnings' contain
			lists of errors and warnings.
		"""
		(n_genes, n_cells) = ds.shape

		self._warn("Description" in ds.attrs, "Optional global attribute 'Description' is missing")
		self._warn("Journal" in ds.attrs, "Optional global attribute 'Journal' is missing")
		self._warn("Authors" in ds.attrs, "Optional global attribute 'Authors' is missing")
		self._warn("Title" in ds.attrs, "Optional global attribute 'Title' is missing")
		self._warn("Year" in ds.attrs, "Optional global attribute 'Year' is missing")
		self._warn("CreationDate" in ds.attrs, "Optional global attribute 'CreationDate' is missing")

		if self._check("ClusterID" in ds.ca, "Column attribute 'ClusterID' is missing"):
			self._check(np.issubdtype(ds.ca.ClusterID.dtype, np.int_), "Column attribute 'ClusterID' must be integer dtype")
			self._check(len(np.unique(ds.ca.ClusterID)) == np.max(ds.ca.ClusterID) and np.min(ds.ca.ClusterID) == 0, "Column attribute 'ClusterID' must be integers 0, 1, 2, ... with no missing values")
			self._check(ds.ca.ClusterID.shape == (n_cells,), f"Column attribute 'ClusterID' must be 1-dimensional array of {n_cells} elements")

		if "ClusterName" in ds.ca:
			self._check(ds.ca.ClusterName.dtype == object and np.issubdtype(ds.ca.ClusterName[0].dtype, np.str_), "Column attribute 'ClusterName' must be an array of strings")
			self._check(ds.ca.ClusterName.shape == (n_cells,), f"Column attribute 'ClusterName' must be 1-dimensional array of {n_cells} elements")
			one_to_one = True
			for cid in np.unique(ds.ca.ClusterID):
				if len(np.unique(ds.ca.ClusterName[ds.ca.ClusterID == cid])) != 1:
					one_to_one = False
					break
			for cn in np.unique(ds.ca.ClusterName):
				if len(np.unique(ds.ca.ClusterID[ds.ca.ClusterName == cn])) != 1:
					one_to_one = False
					break
			if not one_to_one:
				self._check(False, "ClusterName must correspond 1:1 with ClusterID")
		else:
			self.warnings.append("Optional column attribute 'ClusterName' is missing")

		if self._check("CellID" in ds.ca, "Column attribute 'CellID' is missing"):
			self._check(ds.ca.CellID.dtype == object and np.issubdtype(ds.ca.CellID[0].dtype, np.str_), f"Column attribute 'CellID' must be an array of strings, not '{ds.ca.CellID[0].dtype}'")
			self._check(ds.ca.CellID.shape == (n_cells,), f"Column attribute 'CellID' must be 1-dimensional array of {n_cells} elements")
			self._check(len(np.unique(ds.ca.CellID)) == n_cells, "Column attribute 'CellID' cannot contain duplicate values")

		if "Valid" in ds.ca:
			self._check(np.issubdtype(ds.ca.Valid.dtype, np.int_), f"Column attribute 'Valid' must be integer dtype, not '{ds.ca.Valid.dtype}'")
			valids = np.unique(ds.ca.Valid)
			self._check(np.all(np.isin(ds.ca.Valid, [0, 1])), "Column attribute 'Valid' must be integers 0 or 1 only")
			self._check(ds.ca.Valid.shape == (n_cells,), f"Column attribute 'Valid' must be 1-dimensional array of {n_cells} elements")
		else:
			self.warnings.append("Optional column attribute 'Valid' is missing")

		if "Outliers" in ds.ca:
			self._check(np.issubdtype(ds.ca.Outliers.dtype, np.int_), f"Column attribute 'Outliers' must be integer dtype, not '{ds.ca.Outliers.dtype}'")
			self._check(np.all(np.isin(ds.ca.Outliers, [0, 1])), "Column attribute 'Outliers' must be integers 0 or 1 only")
			self._check(ds.ca.Outliers.shape == (n_cells,), f"Column attribute 'Outliers' must be 1-dimensional array of {n_cells} elements")
		else:
			self.warnings.append("Optional column attribute 'Outliers' is missing")

		if self._check("Accession" in ds.ra, "Row attribute 'Accession' is missing"):
			self._check(ds.ra.Accession.dtype == object and np.issubdtype(ds.ra.Accession[0].dtype, np.str_), f"Row attribute 'Accession' must be an array of strings, not '{ds.ra.Accession[0].dtype}'")
			self._check(ds.ra.Accession.shape == (n_genes,), f"Row attribute 'Accession' must be 1-dimensional array of {n_genes} elements")
			self._check(len(np.unique(ds.ra.Accession)) == n_genes, "Row attribute 'Accession' cannot contain duplicate values")

		if self._check("Gene" in ds.ra, "Row attribute 'Gene' is missing"):
			self._check(ds.ra.Gene.dtype == object and np.issubdtype(ds.ra.Gene[0].dtype, np.str_), f"Row attribute 'Gene' must be an array of strings, not '{ds.ra.Gene[0].dtype}'")
			self._check(ds.ra.Gene.shape == (n_genes,), f"Row attribute 'Gene' must be 1-dimensional array of {n_genes} elements")

		if "Valid" in ds.ra:
			self._check(np.issubdtype(ds.ra.Valid.dtype, np.int_), f"Row attribute 'Valid' must be integer dtype, not '{ds.ra.Valid.dtype}'")
			valids = np.unique(ds.ra.Valid)
			self._check(np.all(np.isin(ds.ra.Valid, [0, 1])), "Row attribute 'Valid' must be integers 0 or 1 only")
			self._check(ds.ra.Valid.shape == (n_cells,), f"Row attribute 'Valid' must be 1-dimensional array of {n_cells} elements")
		else:
			self.warnings.append("Optional row attribute 'Valid' is missing")

		if "Selected" in ds.ra:
			self._check(np.issubdtype(ds.ra.Selected.dtype, np.int_), f"Row attribute 'Selected' must be integer dtype, not '{ds.ra.Selected.dtype}'")
			valids = np.unique(ds.ra.Selected)
			self._check(np.all(np.isin(ds.ra.Selected, [0, 1])), "Row attribute 'Selected' must be integers 0 or 1 only")
			self._check(ds.ra.Selected.shape == (n_cells,), f"Row attribute 'Selected' must be 1-dimensional array of {n_cells} elements")
		else:
			self.warnings.append("Optional row attribute 'Selected' is missing")

		return len(self.errors) == 0
		
	def validate_spec(self, file: h5py.File) -> bool:
		"""
		Validate the LoomConnection object against the format specification.

		Args:
			file:			h5py File object
		
		Returns:
			True if the file conforms to the specs, else False
		
		Remarks:
			Upon return, the instance attributes 'self.errors' and 'self.warnings' contain
			lists of errors and warnings, and the 'self.summary' attribute contains a summary
			of the file contents.
		"""
		matrix_types = ["float16", "float32", "float64", "int8", "int16", "int32", "int64", "uint8", "uint16", "uint32", "uint64"]
		vertex_types = ["int8", "int16", "int32", "int64", "uint8", "uint16", "uint32", "uint64"]
		weight_types = ["float16", "float32", "float64"]

		def delay_print(text: str) -> None:
			self.summary.append(text)

		def dt(t: str) -> str:
			if str(t).startswith("|S"):
				return f"string"
			return str(t)

		width_ra = 0
		width_ca = 0
		width_globals = 0
		if self._check("row_attrs" in file, "'row_attrs' group is missing"):
			width_ra = max([len(x) for x in (file["row_attrs"].keys())], default=0)
		if self._check("col_attrs" in file, "'col_attrs' group is missing"):
			width_ca = max([len(x) for x in (file["col_attrs"].keys())], default=0)
		if self.version == "3.0.0":
			if self._check("attrs" in file, "Global attributes missing"):
				width_globals = max([len(x) for x in (file["attrs"].keys())], default=0)
		elif len(file.attrs) > 0:
			width_globals = max([len(x) for x in file.attrs.keys()])
		width_layers = 0
		if "layers" in file and len(file["layers"]) > 0:
			width_layers = max([len(x) for x in file["layers"].keys()])
		width_layers = max(width_layers, len("Main matrix"))
		width = max(width_ca, width_ra, width_globals)

		delay_print("Global attributes:")
		if self.version == "3.0.0":
			self._check("attrs" in file, "Global attributes missing")
			for attr in file["attrs"]:
				if type(attr) is np.ndarray:
					delay_print(f"{attr: >{width}} {attr.dtype} {attr.shape}")
				else:
					delay_print(f"{attr: >{width}} {type(attr).__name__} (scalar)")
		else:
			for key, value in file.attrs.items():
				if type(value) is str:
					self.warnings.append(f"Global attribute '{key}' has dtype string, which will be deprecated in future Loom versions")
					delay_print(f"{key: >{width}} string")
				elif type(value) is bytes:
					self.warnings.append(f"Global attribute '{key}' has dtype bytes, which will be deprecated in future Loom versions")
					delay_print(f"{key: >{width}} bytes")
				else:
					delay_print(f"{key: >{width}} {dt(file.attrs[key].dtype)}")
				
		if self._check("matrix" in file, "Main matrix missing"):
			self._check(file["matrix"].dtype in matrix_types, f"Main matrix dtype={file['matrix'].dtype} is not allowed")
			shape = file["matrix"].shape
			delay_print(f"Layers shape={shape}:")
			delay_print(f"{'Main matrix': >{width}} {file['matrix'].dtype}")

		if "layers" in file:
			for layer in file["layers"]:
				self._check(file["layers"][layer].shape == shape, f"Layer '{layer}' shape {file['layers'][layer].shape} does not match main matrix shape {shape}")
				self._check(file["layers"][layer].dtype in matrix_types, f"Layer '{layer}' dtype={file['layers'][layer].dtype} is not allowed")
				delay_print(f"{layer: >{width}} {file['layers'][layer].dtype}")

		if self.version == "3.0.0":
			expected_dtype = np.object_
		else:
			expected_dtype = np.bytes_
		delay_print("Row attributes:")
		if self._check("row_attrs" in file, "'row_attrs' group is missing"):
			for ra in file["row_attrs"]:
				self._check(file["row_attrs"][ra].shape[0] == shape[0], f"Row attribute '{ra}' shape {file['row_attrs'][ra].shape[0]} first dimension does not match row dimension {shape}")
				self._check(file["row_attrs"][ra].dtype in matrix_types or np.issubdtype(file['row_attrs'][ra].dtype, expected_dtype), f"Row attribute '{ra}' dtype {file['row_attrs'][ra].dtype} is not allowed")
				ra_shape = file['row_attrs'][ra].shape
				delay_print(f"{ra: >{width}} {dt(file['row_attrs'][ra].dtype)} {ra_shape if len(ra_shape) > 1 else ''}")
			if len(file["row_attrs"]) == 0:
				delay_print("    (none)")

		delay_print("Column attributes:")
		if self._check("col_attrs" in file, "'col_attrs' group is missing"):
			for ca in file["col_attrs"]:
				self._check(file["col_attrs"][ca].shape[0] == shape[1], f"Column attribute '{ca}' shape {file['col_attrs'][ca].shape[0]} first dimension does not match column dimension {shape}")
				self._check(file["col_attrs"][ca].dtype in matrix_types or np.issubdtype(file["col_attrs"][ca].dtype, expected_dtype), f"Column attribute '{ca}' dtype {file['col_attrs'][ca].dtype} is not allowed")
				ca_shape = file['col_attrs'][ca].shape
				delay_print(f"{ca: >{width}} {dt(file['col_attrs'][ca].dtype)} {ca_shape if len(ca_shape) > 1 else ''}")
			if len(file["col_attrs"]) == 0:
				delay_print("    (none)")

		delay_print("Row graphs:")
		if "row_graphs" in file:
			if self.version == "2.0.1" or self.version == "3.0.0":
				self._check("row_graphs" in file, "'row_graphs' group is missing (try spec_version='old')")
			for g in file["row_graphs"]:
				self._check("a" in file["row_graphs"][g], f"Row graph '{g}' is missing vector 'a', denoting start vertices")
				self._check(file["row_graphs"][g]['a'].dtype in vertex_types, f"/row_graphs/{g}/a.dtype {file['row_graphs'][g]['a'].dtype} must be integer")
				self._check("b" in file["row_graphs"][g], f"Row graph '{g}' is missing vector 'b', denoting end vertices")
				self._check(file["row_graphs"][g]['b'].dtype in vertex_types, f"/row_graphs/{g}/b.dtype {file['row_graphs'][g]['b'].dtype} must be integer")
				self._check("w" in file["row_graphs"][g], f"Row graph '{g}' is missing vector 'w', denoting vertex weights")
				self._check(file["row_graphs"][g]['w'].dtype in weight_types, f"/row_graphs/{g}/w.dtype {file['row_graphs'][g]['w'].dtype} must be float")
				self._check(file['row_graphs'][g]['a'].shape[0] == file['row_graphs'][g]['b'].shape[0] and file['row_graphs'][g]['a'].shape[0] == file['row_graphs'][g]['w'].shape[0], f"Row graph '{g}' sparse vectors a, b and w must have equal length")
				delay_print(f"    '{g}' with {file['row_graphs'][g]['a'].shape[0]} edges")
			if len(file["row_graphs"]) == 0:
				delay_print("    (none)")

		delay_print("Column graphs:")
		if "col_graphs" in file:
			if self.version == "2.0.1" or self.version == "3.0.0":
				self._check("col_graphs" in file, "'col_graphs' group is missing (try spec_version='old')")
			for g in file["col_graphs"]:
				self._check("a" in file["col_graphs"][g], f"Column graph '{g}' is missing vector 'a', denoting start vertices")
				self._check(file["col_graphs"][g]['a'].dtype in vertex_types, f"/col_graphs/{g}/a.dtype {file['col_graphs'][g]['a'].dtype} must be integer")
				self._check("b" in file["col_graphs"][g], f"Column graph '{g}' is missing vector 'b', denoting end vertices")
				self._check(file["col_graphs"][g]['b'].dtype in vertex_types, f"/col_graphs/{g}/b.dtype {file['col_graphs'][g]['b'].dtype} must be integer")
				self._check("w" in file["col_graphs"][g], f"Column graph '{g}' is missing vector 'w', denoting vertex weights")
				self._check(file["col_graphs"][g]['w'].dtype in weight_types, f"/col_graphs/{g}/w.dtype {file['col_graphs'][g]['w'].dtype} must be float")
				self._check(file['col_graphs'][g]['a'].shape[0] == file['col_graphs'][g]['b'].shape[0] and file['col_graphs'][g]['a'].shape[0] == file['col_graphs'][g]['w'].shape[0], f"Column graph '{g}' sparse vectors a, b and w must have equal length")
				delay_print(f"    '{g}' with {file['col_graphs'][g]['a'].shape[0]} edges")
			if len(file["col_graphs"]) == 0:
				delay_print("    (none)")

		return len(self.errors) == 0
